@@ -1,3 +1,4 @@
+# coding: utf8
 from __future__ import print_function
 import logging
 import sys
@@ -30,6 +31,8 @@ from fresco.util.urls import url_join, make_query
 from fresco.util.http import parse_header
 from fresco.util.wsgi import unicode_to_environ
 
+__version__ = '7.4.1'
+
 #: Registry for xpath multimethods
 xpath_registry = {}
 
@@ -1006,17 +1009,31 @@ class TestAgent(object):
         assert not re.search(r'%(?![A-F0-9]{2})', PATH_INFO), \
             "Path info not URL encoded"
 
+        # Unquote requires a string argument
+        if isinstance(PATH_INFO, bytes):
+            PATH_INFO = PATH_INFO.decode('ascii')
+        if isinstance(SCRIPT_NAME, bytes):
+            SCRIPT_NAME = SCRIPT_NAME.decode('ascii')
+
+        # 'caf%C3%A9' -> 'caf\xc3\xa9'
         PATH_INFO = unquote(PATH_INFO)
         SCRIPT_NAME = unquote(SCRIPT_NAME)
+
+        # Python 3 unquotes to unicode correctly but Python 2 gets it wrong
+        if not PY3:
+            # 'caf\xc3\xa0' -> b'caf\xc3\xa9' -> 'café'
+            PATH_INFO = PATH_INFO.encode('latin1').decode(self.charset)
+            SCRIPT_NAME = SCRIPT_NAME.encode('latin1').decode(self.charset)
+
         PATH_INFO = unicode_to_environ(PATH_INFO)
         SCRIPT_NAME = unicode_to_environ(SCRIPT_NAME)
 
         environ = self.environ_defaults.copy()
         environ.update(kwargs)
         for key, value in kwargs.items():
-            environ[key.replace('wsgi_', 'wsgi.')] = value
             if '.' not in key:
                 value = unicode_to_environ(value)
+            environ[key.replace('wsgi_', 'wsgi.')] = value
 
         if isinstance(wsgi_input, bytes):
             wsgi_input = BytesIO(wsgi_input)
@@ -1640,7 +1657,7 @@ class TestAgent(object):
         port = int(self.environ_defaults['SERVER_PORT'])
 
         server = make_server(host, port, PassStateWSGIApp(self))
-        url = self.request.make_uri(netloc='{0}:{1}'.format(host, port))
+        url = self.request.make_url(netloc='{0}:{1}'.format(host, port))
         if open_in_browser:
             webbrowser.open_new_tab(url)
         print("\nStarting HTTP server on {0}:{1}. "
@@ -1687,9 +1704,12 @@ class PassStateWSGIApp(object):
         response = self.testagent.response
         mimetype, charset = parse_header(response.get_header('Content-Type'))
         charset = dict(charset).get('charset', 'UTF-8')
+        self.testagent._read_response()
         if is_html(response) and self.testagent._lxml is not None:
             response = response.replace(content=[tostring(self.testagent.lxml,
                                                           encoding=charset)])
+        else:
+            response = response.replace(content=[self.testagent._body])
 
         for key, morsel in self.testagent.cookies.items():
             response = response.add_header('Set-Cookie', morsel.OutputString())
